@@ -1,19 +1,19 @@
 /* The MIT License
 
-   Copyright (c) 2014-2021 Genome Research Ltd.
+   Copyright (c) 2014-2024 Genome Research Ltd.
 
    Author: Petr Danecek <pd3@sanger.ac.uk>
-   
+
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:
-   
+
    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.
-   
+
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
@@ -42,6 +42,7 @@
 #include "bcftools.h"
 #include "rbuf.h"
 #include "filter.h"
+#include "smpl_ilist.h"
 
 // Logic of the filters: include or exclude sites which match the filters?
 #define FLT_INCLUDE 1
@@ -53,8 +54,8 @@
 #define PICK_SHORT 8
 #define PICK_IUPAC 16
 
-#define TO_UPPER 0
-#define TO_LOWER 1
+#define TO_UPPER 1
+#define TO_LOWER 2
 
 typedef struct
 {
@@ -115,17 +116,26 @@ typedef struct
     FILE *fp_out;
     FILE *fp_chain;
     char **argv;
-    int argc, output_iupac, haplotype, allele, isample, napplied;
-    uint8_t *iupac_bitmask;
-    int miupac_bitmask;
-    char *fname, *ref_fname, *sample, *output_fname, *mask_fname, *chain_fname, missing_allele, absent_allele;
+    int argc, output_iupac, iupac_GTs, haplotype, allele, isample, napplied;
+    uint8_t *iupac_bitmask, *iupac_als;
+    int miupac_bitmask, miupac_als, regions_overlap;
+    char *fname, *ref_fname, *sample, *sample_fname, *output_fname, *mask_fname, *chain_fname, missing_allele, absent_allele;
     char mark_del, mark_ins, mark_snv;
+    smpl_ilist_t *smpl;
 }
 args_t;
 
+static void destroy_chain(chain_t *chain)
+{
+    if ( !chain ) return;
+    free(chain->ref_gaps);
+    free(chain->alt_gaps);
+    free(chain->block_lengths);
+    free(chain);
+}
 static chain_t* init_chain(chain_t *chain, int ref_ori_pos)
 {
-//     fprintf(stderr, "init_chain(*chain, ref_ori_pos=%d)\n", ref_ori_pos);
+    if ( chain ) destroy_chain(chain);
     chain = (chain_t*) calloc(1,sizeof(chain_t));
     chain->num = 0;
     chain->block_lengths = NULL;
@@ -137,18 +147,6 @@ static chain_t* init_chain(chain_t *chain, int ref_ori_pos)
     return chain;
 }
 
-static void destroy_chain(args_t *args)
-{
-    chain_t *chain = args->chain;
-    free(chain->ref_gaps);
-    free(chain->alt_gaps);
-    free(chain->block_lengths);
-    free(chain);
-    chain = NULL;
-    free(args->chr);
-    args->chr = NULL;
-}
-
 static void print_chain(args_t *args)
 {
     /*
@@ -172,7 +170,7 @@ static void print_chain(args_t *args)
         - alt_start (same as ref_start, as no edits are recorded/applied before that position)
         - alt_end (adjusted to match the length of the alt sequence)
         - chain_num (just an auto-increment id)
-        
+
         the other (sorted) lines are:
         - length of the ungapped alignment block
         - gap on the ref sequence between this and the next block (all but the last line)
@@ -197,7 +195,7 @@ static void print_chain(args_t *args)
 
 static void push_chain_gap(chain_t *chain, int ref_start, int ref_len, int alt_start, int alt_len)
 {
-//     fprintf(stderr, "push_chain_gap(*chain, ref_start=%d, ref_len=%d, alt_start=%d, alt_len=%d)\n", ref_start, ref_len, alt_start, alt_len);
+    // fprintf(stderr, "push_chain_gap(chain=%p, ref_start=%d, ref_len=%d, alt_start=%d, alt_len=%d)\n", chain, ref_start, ref_len, alt_start, alt_len);
     int num = chain->num;
 
     if (num && ref_start <= chain->ref_last_block_ori) {
@@ -230,15 +228,27 @@ static void init_data(args_t *args)
     if ( !bcf_sr_add_reader(args->files,args->fname) ) error("Failed to read from %s: %s\n", !strcmp("-",args->fname)?"standard input":args->fname, bcf_sr_strerror(args->files->errnum));
     args->hdr = args->files->readers[0].header;
     args->isample = -1;
-    if ( args->sample )
+    if ( !args->sample )
+        args->smpl = smpl_ilist_init(args->hdr,NULL,0,SMPL_NONE|SMPL_VERBOSE);
+    else if ( args->sample && strcmp("-",args->sample) )
     {
-        args->isample = bcf_hdr_id2int(args->hdr,BCF_DT_SAMPLE,args->sample);
-        if ( args->isample<0 ) error("No such sample: %s\n", args->sample);
+        args->smpl = smpl_ilist_init(args->hdr,args->sample,0,SMPL_NONE|SMPL_VERBOSE);
+        if ( args->smpl && !args->smpl->n ) error("No matching sample found\n");
     }
-    if ( (args->haplotype || args->allele) && args->isample<0 )
+    else if ( args->sample_fname )
     {
-        if ( bcf_hdr_nsamples(args->hdr) > 1 ) error("The --sample option is expected with --haplotype\n");
-        args->isample = 0;
+        args->smpl = smpl_ilist_init(args->hdr,args->sample_fname,1,SMPL_NONE|SMPL_VERBOSE);
+        if ( args->smpl && !args->smpl->n ) error("No matching sample found\n");
+    }
+    if ( args->smpl )
+    {
+        if ( args->haplotype || args->allele )
+        {
+            if ( args->smpl->n > 1 ) error("Too many samples, only one can be used with -H\n");
+            args->isample = args->smpl->idx[0];
+        }
+        else
+            args->iupac_GTs = 1;
     }
     int i;
     for (i=0; i<args->nmask; i++)
@@ -262,7 +272,7 @@ static void init_data(args_t *args)
         if ( ! args->fp_out ) error("Failed to create %s: %s\n", args->output_fname, strerror(errno));
     }
     else args->fp_out = stdout;
-    if ( args->isample<0 ) fprintf(stderr,"Note: the --sample option not given, applying all records regardless of the genotype\n");
+    if ( args->isample<0 && !args->iupac_GTs ) fprintf(stderr,"Note: the --samples option not given, applying all records regardless of the genotype\n");
     if ( args->filter_str )
         args->filter = filter_init(args->hdr, args->filter_str);
     args->rid = -1;
@@ -286,8 +296,10 @@ static void add_mask_with(args_t *args, char *with)
 }
 static void destroy_data(args_t *args)
 {
+    free(args->iupac_als);
     free(args->iupac_bitmask);
     if (args->filter) filter_destroy(args->filter);
+    if ( args->smpl ) smpl_ilist_destroy(args->smpl);
     bcf_sr_destroy(args->files);
     int i;
     for (i=0; i<args->vcf_rbuf.m; i++)
@@ -305,13 +317,14 @@ static void destroy_data(args_t *args)
     if ( args->chain_fname )
         if ( fclose(args->fp_chain) ) error("Close failed: %s\n", args->chain_fname);
     if ( fclose(args->fp_out) ) error("Close failed: %s\n", args->output_fname);
+    destroy_chain(args->chain);
 }
 
 static void init_region(args_t *args, char *line)
 {
     char *ss, *se = line;
     while ( *se && !isspace(*se) && *se!=':' ) se++;
-    int from = 0, to = 0;
+    hts_pos_t from = 0, to = 0;
     char tmp = 0, *tmp_ptr = NULL;
     if ( *se )
     {
@@ -335,23 +348,34 @@ static void init_region(args_t *args, char *line)
     args->prev_base_pos = -1;
     args->fa_buf.l  = 0;
     args->fa_length = 0;
-    args->fa_end_pos = to;
-    args->fa_ori_pos = from;
-    args->fa_src_pos = from;
+    args->fa_end_pos = to;      // 0-based
+    args->fa_ori_pos = from;    // 0-based
+    args->fa_src_pos = from;    // 0-based
     args->fa_mod_off = 0;
     args->fa_frz_pos = -1;
     args->fa_frz_mod = -1;
     args->fa_case    = -1;
     args->vcf_rbuf.n = 0;
-    bcf_sr_seek(args->files,line,args->fa_ori_pos);
+
+
+    // bcf_sr_set_regions accepts 1-based coordinates
+    kstring_t str = {0,0,0};
+    if ( !from ) from = 1;
+    else from++;
+#ifndef MAX_CSI_COOR
+#define MAX_CSI_COOR ((1LL << (14 + 30)) - 1)
+#endif
+    if ( to==0 ) to = MAX_CSI_COOR - 1;
+    else to++;
+    ksprintf(&str,"%s:%"PRIhts_pos"-%"PRIhts_pos,line,from,to);
+    bcf_sr_set_opt(args->files,BCF_SR_REGIONS_OVERLAP,args->regions_overlap);
+    bcf_sr_set_regions(args->files,str.s,0);
+    free(str.s);
+
     if ( tmp_ptr ) *tmp_ptr = tmp;
     fprintf(args->fp_out,">%s%s\n",args->chr_prefix?args->chr_prefix:"",line);
-    if (args->chain_fname )
-    {
+    if ( args->chain_fname )
         args->chain = init_chain(args->chain, args->fa_ori_pos);
-    } else {
-        args->chain = NULL;
-    }
 }
 
 static bcf1_t **next_vcf_line(args_t *args)
@@ -457,25 +481,90 @@ static char *mark_del(char *ref, int rlen, char *alt, int mark)
 static void mark_ins(char *ref, char *alt, char mark)
 {
     int i, nref = strlen(ref), nalt = strlen(alt);
-    if ( mark=='l' )
+    if ( mark==TO_LOWER )
         for (i=nref; i<nalt; i++) alt[i] = tolower(alt[i]);
-    else
+    else if ( mark==TO_UPPER )
         for (i=nref; i<nalt; i++) alt[i] = toupper(alt[i]);
+    else if ( mark )
+        for (i=nref; i<nalt; i++) alt[i] = mark;
 }
 static void mark_snv(char *ref, char *alt, char mark)
 {
     int i, nref = strlen(ref), nalt = strlen(alt);
     int n = nref < nalt ? nref : nalt;
-    if ( mark=='l' )
+    if ( mark==TO_LOWER )
     {
         for (i=0; i<n; i++)
             if ( tolower(ref[i])!=tolower(alt[i]) ) alt[i] = tolower(alt[i]);
     }
-    else
+    else if ( mark==TO_UPPER)
     {
         for (i=0; i<n; i++)
             if ( tolower(ref[i])!=tolower(alt[i]) ) alt[i] = toupper(alt[i]);
     }
+    else if ( mark==TO_UPPER)
+    {
+        for (i=0; i<n; i++)
+            if ( tolower(ref[i])!=tolower(alt[i]) ) alt[i] = toupper(alt[i]);
+    }
+    else if ( mark )
+    {
+        for (i=0; i<n; i++)
+            if ( tolower(ref[i])!=tolower(alt[i]) ) alt[i] = mark;
+    }
+}
+static void iupac_init(args_t *args, bcf1_t *rec)
+{
+    int i;
+    hts_resize(uint8_t, rec->n_allele, &args->miupac_als, &args->iupac_als, 0);
+    for (i=0; i<args->miupac_als; i++) args->iupac_als[i] = 0;
+}
+static int iupac_add_gt(args_t *args, bcf1_t *rec, uint8_t *gt, int ngt)
+{
+    int i, is_set = 0;
+    for (i=0; i<ngt; i++)
+    {
+        if ( bcf_gt_is_missing(gt[i]) ) continue;
+        if ( gt[i]==(uint8_t)bcf_int8_vector_end ) break;
+        int ial = bcf_gt_allele(gt[i]);
+        if ( ial >= rec->n_allele ) error("Invalid VCF, too few ALT alleles at %s:%"PRId64"\n", bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1);
+        args->iupac_als[ial] = 1;
+        is_set = 1;
+    }
+    return is_set;
+}
+static int iupac_set_allele(args_t *args, bcf1_t *rec)
+{
+    int i,j, max_len = 0, alt_len = 0, ialt = -1, fallback_alt = -1;
+    for (i=0; i<rec->n_allele; i++)
+    {
+        if ( !args->iupac_als[i] ) continue;
+        if ( fallback_alt <=0 ) fallback_alt = i;
+        int l = strlen(rec->d.allele[i]);
+        for (j=0; j<l; j++)
+            if ( iupac2bitmask(rec->d.allele[i][j]) < 0 ) break;
+        if ( j<l ) continue; // symbolic allele, breakpoint or invalid character in the allele
+        if ( l>max_len )
+        {
+            hts_resize(uint8_t, l, &args->miupac_bitmask, &args->iupac_bitmask, HTS_RESIZE_CLEAR);
+            for (j=max_len; j<l; j++) args->iupac_bitmask[j] = 0;
+            max_len = l;
+        }
+        if ( i>0 && l>alt_len )
+        {
+            alt_len = l;
+            ialt = i;
+        }
+        for (j=0; j<l; j++)
+            args->iupac_bitmask[j] |= iupac2bitmask(rec->d.allele[i][j]);
+    }
+    if ( alt_len > 0 )
+    {
+        for (j=0; j<alt_len; j++) rec->d.allele[ialt][j] = bitmask2iupac(args->iupac_bitmask[j]);
+        return ialt;
+    }
+    if ( fallback_alt >= 0 ) return fallback_alt;
+    return ialt;
 }
 static void apply_variant(args_t *args, bcf1_t *rec)
 {
@@ -498,7 +587,25 @@ static void apply_variant(args_t *args, bcf1_t *rec)
     }
 
     int ialt = 1;    // the alternate allele
-    if ( args->isample >= 0 )
+    if ( args->iupac_GTs )
+    {
+        bcf_unpack(rec, BCF_UN_FMT);
+        bcf_fmt_t *fmt = bcf_get_fmt(args->hdr, rec, "GT");
+        if ( !fmt ) return;
+        if ( fmt->type!=BCF_BT_INT8 )
+            error("Todo: GT field represented with BCF_BT_INT8, too many alleles at %s:%"PRId64"?\n",bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1);
+        ialt = -1;
+        int is_set = 0;
+        iupac_init(args,rec);
+        for (i=0; i<args->smpl->n; i++)
+        {
+            uint8_t *ptr = fmt->p + fmt->size*args->smpl->idx[i];
+            is_set += iupac_add_gt(args, rec, ptr, fmt->n);
+        }
+        if ( !is_set && !args->missing_allele ) return;
+        if ( is_set ) ialt = iupac_set_allele(args, rec);
+    }
+    else if ( args->isample >= 0 )
     {
         bcf_unpack(rec, BCF_UN_FMT);
         bcf_fmt_t *fmt = bcf_get_fmt(args->hdr, rec, "GT");
@@ -526,7 +633,7 @@ static void apply_variant(args_t *args, bcf1_t *rec)
                     if ( !args->missing_allele ) return;
                     ialt = -1;
                 }
-                else 
+                else
                 {
                     if ( !warned_haplotype )
                     {
@@ -544,46 +651,17 @@ static void apply_variant(args_t *args, bcf1_t *rec)
                     if ( !args->missing_allele ) return;
                     ialt = -1;
                 }
-                else 
+                else
                     ialt = bcf_gt_allele(ialt);
             }
         }
-        else if ( action==use_iupac ) 
+        else if ( action==use_iupac )
         {
             ialt = -1;
-            int is_missing = 0, alen = 0, mlen = 0, fallback_alt = -1;
-            for (i=0; i<fmt->n; i++)
-            {
-                if ( bcf_gt_is_missing(ptr[i]) ) { is_missing = 1; continue; }
-                if ( ptr[i]==(uint8_t)bcf_int8_vector_end ) break;
-                int jalt = bcf_gt_allele(ptr[i]);
-                if ( jalt >= rec->n_allele ) error("Invalid VCF, too few ALT alleles at %s:%"PRId64"\n", bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1);
-                if ( fallback_alt <= 0 ) fallback_alt = jalt;
-
-                int l = strlen(rec->d.allele[jalt]);
-                for (j=0; j<l; j++)
-                    if ( iupac2bitmask(rec->d.allele[jalt][j]) < 0 ) break;
-                if ( j<l ) continue; // symbolic allele, breakpoint or invalid character in the allele
-
-                if ( l > mlen )
-                {
-                    hts_expand(uint8_t,l,args->miupac_bitmask,args->iupac_bitmask);
-                    for (j=mlen; j<l; j++) args->iupac_bitmask[j] = 0;
-                    mlen = l;
-                }
-                if ( jalt>0 && l>alen )
-                {
-                    alen = l;
-                    ialt = jalt;
-                }
-                for (j=0; j<l; j++)
-                    args->iupac_bitmask[j] |= iupac2bitmask(rec->d.allele[jalt][j]);
-            }
-            if ( alen > 0 )
-                for (j=0; j<alen; j++) rec->d.allele[ialt][j] = bitmask2iupac(args->iupac_bitmask[j]);
-            else if ( fallback_alt >= 0 )
-                ialt = fallback_alt;
-            else if ( is_missing && !args->missing_allele ) return;
+            iupac_init(args,rec);
+            int is_set = iupac_add_gt(args, rec, ptr, fmt->n);
+            if ( !is_set && !args->missing_allele ) return;
+            if ( is_set ) ialt = iupac_set_allele(args, rec);
         }
         else
         {
@@ -717,18 +795,45 @@ static void apply_variant(args_t *args, bcf1_t *rec)
             fprintf(stderr,"The site %s:%"PRId64" overlaps with another variant, skipping...\n", bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1);
             return;
         }
-        
+
     }
 
+    char *ref_allele = rec->d.allele[0];
     char *alt_allele = rec->d.allele[ialt];
     int rmme_alt = 0;
 
     int len_diff = 0, alen = 0;
-    int idx = rec->pos - args->fa_ori_pos + args->fa_mod_off;
+    int idx = rec->pos - args->fa_ori_pos + args->fa_mod_off;   // position of the variant within the modified fasta sequence
     if ( idx<0 )
     {
-        fprintf(stderr,"Warning: ignoring overlapping variant starting at %s:%"PRId64"\n", bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1);
-        return;
+        if ( alt_allele[0]=='<' )   // symbolic allele
+        {
+            rec->pos  -= idx + 1;
+            rec->rlen += idx + 1;
+            idx = -1;
+        }
+        else if ( strlen(ref_allele) < -idx )   // the ref allele is shorter but overlaps the fa sequence? This should never happen
+        {
+            assert(0);
+            fprintf(stderr,"Warning: ignoring overlapping variant starting at %s:%"PRId64"\n", bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1);
+            return;
+        }
+        else if ( strlen(alt_allele) > -idx )  // the ref allele overlaps the fa and so does the alt allele
+        {
+            rec->pos   -= idx;
+            rec->rlen  += idx;
+            ref_allele -= idx;
+            alt_allele -= idx;
+            idx = 0;
+        }
+        else    // the ref allele overlaps the fa but alt allele does not: trim to leave one base before
+        {
+            rec->pos   -= idx + 1;
+            rec->rlen  += idx + 1;
+            ref_allele -= idx + 1;
+            alt_allele += strlen(alt_allele) - 1;
+            idx = -1;
+        }
     }
     if ( rec->rlen > args->fa_buf.l - idx )
     {
@@ -743,7 +848,7 @@ static void apply_variant(args_t *args, bcf1_t *rec)
             }
         }
     }
-    if ( idx>=args->fa_buf.l ) 
+    if ( idx>0 && idx>=args->fa_buf.l )
         error("FIXME: %s:%"PRId64" .. idx=%d, ori_pos=%d, len=%"PRIu64", off=%d\n",bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1,idx,args->fa_ori_pos,(uint64_t)args->fa_buf.l,args->fa_mod_off);
 
     // sanity check the reference base
@@ -758,7 +863,7 @@ static void apply_variant(args_t *args, bcf1_t *rec)
         if ( !strcasecmp(alt_allele,"<DEL>") )
         {
             static int multibase_ref_del_warned = 0;
-            if ( rec->d.allele[0][1]!=0 && !multibase_ref_del_warned )
+            if ( ref_allele[1]!=0 && !multibase_ref_del_warned )
             {
                 fprintf(stderr,
                     "Warning: one REF base is expected with <DEL>, assuming the actual deletion starts at POS+1 at %s:%"PRId64".\n"
@@ -767,7 +872,7 @@ static void apply_variant(args_t *args, bcf1_t *rec)
             }
             if ( args->mark_del )   // insert dashes instead of delete sequence
             {
-                alt_allele = mark_del(rec->d.allele[0], rec->rlen, NULL, args->mark_del);
+                alt_allele = mark_del(ref_allele, rec->rlen, NULL, args->mark_del);
                 alen = rec->rlen;
                 len_diff = 0;
                 rmme_alt = 1;
@@ -775,7 +880,7 @@ static void apply_variant(args_t *args, bcf1_t *rec)
             else
             {
                 len_diff = 1-rec->rlen;
-                alt_allele = rec->d.allele[0];     // according to VCF spec, the first REF base must precede the event
+                alt_allele = ref_allele;     // according to VCF spec, the first REF base must precede the event
                 alen = 1;
             }
         }
@@ -786,7 +891,7 @@ static void apply_variant(args_t *args, bcf1_t *rec)
             return;
         }
     }
-    else if ( strncasecmp(rec->d.allele[0],args->fa_buf.s+idx,rec->rlen) )
+    else if ( idx>=0 && strncasecmp(ref_allele,args->fa_buf.s+idx,rec->rlen) )
     {
         // This is hacky, handle a special case: if SNP or an insert follows a deletion (AAC>A, C>CAA),
         // the reference base in fa_buf is lost and the check fails. We do not keep a buffer
@@ -794,17 +899,17 @@ static void apply_variant(args_t *args, bcf1_t *rec)
         // one base overlap
 
         int fail = 1;
-        if ( args->prev_base_pos==rec->pos && toupper(rec->d.allele[0][0])==toupper(args->prev_base) )
+        if ( args->prev_base_pos==rec->pos && toupper(ref_allele[0])==toupper(args->prev_base) )
         {
             if ( rec->rlen==1 ) fail = 0;
-            else if ( !strncasecmp(rec->d.allele[0]+1,args->fa_buf.s+idx+1,rec->rlen-1) ) fail = 0;
+            else if ( !strncasecmp(ref_allele+1,args->fa_buf.s+idx+1,rec->rlen-1) ) fail = 0;
         }
 
         if ( fail )
         {
             char tmp = 0;
-            if ( args->fa_buf.l - idx > rec->rlen ) 
-            { 
+            if ( args->fa_buf.l - idx > rec->rlen )
+            {
                 tmp = args->fa_buf.s[idx+rec->rlen];
                 args->fa_buf.s[idx+rec->rlen] = 0;
             }
@@ -813,16 +918,16 @@ static void apply_variant(args_t *args, bcf1_t *rec)
                     "   REF .vcf: [%s]\n"
                     "   ALT .vcf: [%s]\n"
                     "   REF .fa : [%s]%c%s\n",
-                    bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1, rec->d.allele[0], alt_allele, args->fa_buf.s+idx,
+                    bcf_seqname(args->hdr,rec),(int64_t) rec->pos+1, ref_allele, alt_allele, args->fa_buf.s+idx,
                     tmp?tmp:' ',tmp?args->fa_buf.s+idx+rec->rlen+1:""
                  );
         }
         alen = strlen(alt_allele);
         len_diff = alen - rec->rlen;
 
-        if ( args->mark_del && len_diff<0 ) 
+        if ( args->mark_del && len_diff<0 )
         {
-            alt_allele = mark_del(rec->d.allele[0], rec->rlen, alt_allele, args->mark_del);
+            alt_allele = mark_del(ref_allele, rec->rlen, alt_allele, args->mark_del);
             alen = rec->rlen;
             len_diff = 0;
             rmme_alt = 1;
@@ -833,25 +938,26 @@ static void apply_variant(args_t *args, bcf1_t *rec)
         alen = strlen(alt_allele);
         len_diff = alen - rec->rlen;
 
-        if ( args->mark_del && len_diff<0 ) 
+        if ( args->mark_del && len_diff<0 )
         {
-            alt_allele = mark_del(rec->d.allele[0], rec->rlen, alt_allele, args->mark_del);
+            alt_allele = mark_del(ref_allele, rec->rlen, alt_allele, args->mark_del);
             alen = rec->rlen;
             len_diff = 0;
             rmme_alt = 1;
         }
     }
 
-    args->fa_case = toupper(args->fa_buf.s[idx])==args->fa_buf.s[idx] ? TO_UPPER : TO_LOWER;
+    int safe_idx = idx<0 ? 0 : idx; // idx can be negative in case of overlapping deletion
+    args->fa_case = toupper(args->fa_buf.s[safe_idx])==args->fa_buf.s[safe_idx] ? TO_UPPER : TO_LOWER;
     if ( args->fa_case==TO_UPPER )
         for (i=0; i<alen; i++) alt_allele[i] = toupper(alt_allele[i]);
     else
         for (i=0; i<alen; i++) alt_allele[i] = tolower(alt_allele[i]);
 
     if ( args->mark_ins && len_diff>0 )
-        mark_ins(rec->d.allele[0], alt_allele, args->mark_ins);
+        mark_ins(ref_allele, alt_allele, args->mark_ins);
     if ( args->mark_snv )
-        mark_snv(rec->d.allele[0], alt_allele, args->mark_snv);
+        mark_snv(ref_allele, alt_allele, args->mark_snv);
 
     if ( len_diff <= 0 )
     {
@@ -883,7 +989,7 @@ static void apply_variant(args_t *args, bcf1_t *rec)
         //      1   C   T
         //      1   C   CAA
         int ibeg = 0;
-        while ( ibeg<alen && rec->d.allele[0][ibeg]==alt_allele[ibeg] && rec->pos + ibeg <= args->prev_base_pos  ) ibeg++;
+        while ( ibeg<alen && ref_allele[ibeg]==alt_allele[ibeg] && rec->pos + ibeg <= args->prev_base_pos  ) ibeg++;
         for (i=ibeg; i<alen; i++)
             args->fa_buf.s[idx+i] = alt_allele[i];
 
@@ -892,7 +998,7 @@ static void apply_variant(args_t *args, bcf1_t *rec)
     if (args->chain && len_diff != 0)
     {
         // If first nucleotide of both REF and ALT are the same... (indels typically include the nucleotide before the variant)
-        if ( strncasecmp(rec->d.allele[0],alt_allele,1) == 0)
+        if ( strncasecmp(ref_allele,alt_allele,1) == 0)
         {
             // ...extend the block by 1 bp: start is 1 bp further and alleles are 1 bp shorter
             push_chain_gap(args->chain, rec->pos + 1, rec->rlen - 1, rec->pos + 1 + args->fa_mod_off, alen - 1);
@@ -949,10 +1055,8 @@ static void consensus(args_t *args)
         if ( str.s[0]=='>' )
         {
             // new sequence encountered
-            if (args->chain) {
-                print_chain(args);
-                destroy_chain(args);
-            }
+            if ( args->chain ) print_chain(args);
+
             // apply all cached variants and variants that might have been missed because of short fasta (see test/consensus.9.*)
             bcf1_t **rec_ptr = NULL;
             while ( args->rid>=0 && (rec_ptr = next_vcf_line(args)) )
@@ -1026,11 +1130,7 @@ static void consensus(args_t *args)
         if ( args->fa_ori_pos + args->fa_buf.l - args->fa_mod_off <= rec->pos ) break;
         apply_variant(args, rec);
     }
-    if (args->chain)
-    {
-        print_chain(args);
-        destroy_chain(args);
-    }
+    if (args->chain) print_chain(args);
     if ( args->absent_allele ) apply_absent(args, HTS_POS_MAX);
     flush_fa_buffer(args, 0);
     bgzf_close(fasta);
@@ -1043,41 +1143,44 @@ static void usage(args_t *args)
     fprintf(stderr, "\n");
     fprintf(stderr, "About: Create consensus sequence by applying VCF variants to a reference fasta\n");
     fprintf(stderr, "       file. By default, the program will apply all ALT variants. Using the\n");
-    fprintf(stderr, "       --sample (and, optionally, --haplotype) option will apply genotype\n");
+    fprintf(stderr, "       --samples (and, optionally, --haplotype) option will apply genotype\n");
     fprintf(stderr, "       (or haplotype) calls from FORMAT/GT. The program ignores allelic depth\n");
     fprintf(stderr, "       information, such as INFO/AD or FORMAT/AD.\n");
     fprintf(stderr, "Usage: bcftools consensus [OPTIONS] <file.vcf.gz>\n");
     fprintf(stderr, "Options:\n");
-    fprintf(stderr, "    -c, --chain FILE               write a chain file for liftover\n");
-    fprintf(stderr, "    -a, --absent CHAR              replace positions absent from VCF with CHAR\n");
-    fprintf(stderr, "    -e, --exclude EXPR             exclude sites for which the expression is true (see man page for details)\n");
-    fprintf(stderr, "    -f, --fasta-ref FILE           reference sequence in fasta format\n");
-    fprintf(stderr, "    -H, --haplotype WHICH          choose which allele to use from the FORMAT/GT field, note\n");
+    fprintf(stderr, "    -c, --chain FILE               Write a chain file for liftover\n");
+    fprintf(stderr, "    -a, --absent CHAR              Replace positions absent from VCF with CHAR\n");
+    fprintf(stderr, "    -e, --exclude EXPR             Exclude sites for which the expression is true (see man page for details)\n");
+    fprintf(stderr, "    -f, --fasta-ref FILE           Reference sequence in fasta format\n");
+    fprintf(stderr, "    -H, --haplotype WHICH          Choose which allele to use from the FORMAT/GT field, note\n");
     fprintf(stderr, "                                   the codes are case-insensitive:\n");
-    fprintf(stderr, "                                       1: first allele from GT, regardless of phasing\n");
-    fprintf(stderr, "                                       2: second allele from GT, regardless of phasing\n");
+    fprintf(stderr, "                                       N: N={1,2,3,..} is the index of the allele from GT, regardless of phasing (e.g. \"2\")\n");
     fprintf(stderr, "                                       R: REF allele in het genotypes\n");
     fprintf(stderr, "                                       A: ALT allele\n");
     fprintf(stderr, "                                       I: IUPAC code for all genotypes\n");
     fprintf(stderr, "                                       LR,LA: longer allele and REF/ALT if equal length\n");
     fprintf(stderr, "                                       SR,SA: shorter allele and REF/ALT if equal length\n");
-    fprintf(stderr, "                                       1pIu,2pIu: first/second allele for phased and IUPAC code for unphased GTs\n");
-    fprintf(stderr, "    -i, --include EXPR             select sites for which the expression is true (see man page for details)\n");
-    fprintf(stderr, "    -I, --iupac-codes              output variants in the form of IUPAC ambiguity codes\n");
-    fprintf(stderr, "        --mark-del CHAR            instead of removing sequence, insert CHAR for deletions\n");
-    fprintf(stderr, "        --mark-ins uc|lc           highlight insertions in uppercase (uc) or lowercase (lc), leaving the rest as is\n");
-    fprintf(stderr, "        --mark-snv uc|lc           highlight substitutions in uppercase (uc) or lowercase (lc), leaving the rest as is\n");
-    fprintf(stderr, "    -m, --mask FILE                replace regions according to the next --mask-with option. The default is --mask-with N\n");
-    fprintf(stderr, "        --mask-with CHAR|uc|lc     replace with CHAR (skips overlapping variants); change to uppercase (uc) or lowercase (lc)\n");
-    fprintf(stderr, "    -M, --missing CHAR             output CHAR instead of skipping a missing genotype \"./.\"\n");
-    fprintf(stderr, "    -o, --output FILE              write output to a file [standard output]\n");
-    fprintf(stderr, "    -p, --prefix STRING            prefix to add to output sequence names\n");
-    fprintf(stderr, "    -s, --sample NAME              apply variants of the given sample\n");
+    fprintf(stderr, "                                       NpIu: index of the allele for phased and IUPAC code for unphased GTs (e.g. \"2pIu\")\n");
+    fprintf(stderr, "    -i, --include EXPR             Select sites for which the expression is true (see man page for details)\n");
+    fprintf(stderr, "    -I, --iupac-codes              Output IUPAC codes based on FORMAT/GT, use -s/-S to subset samples\n");
+    fprintf(stderr, "        --mark-del CHAR            Instead of removing sequence, insert character CHAR for deletions\n");
+    fprintf(stderr, "        --mark-ins uc|lc|CHAR      Highlight insertions in uppercase (uc), lowercase (lc), or use CHAR, leaving the rest as is\n");
+    fprintf(stderr, "        --mark-snv uc|lc|CHAR      Highlight substitutions in uppercase (uc), lowercase (lc), or use CHAR, leaving the rest as is\n");
+    fprintf(stderr, "    -m, --mask FILE                Replace regions according to the next --mask-with option. The default is --mask-with N\n");
+    fprintf(stderr, "        --mask-with CHAR|uc|lc     Replace with CHAR (skips overlapping variants); change to uppercase (uc) or lowercase (lc)\n");
+    fprintf(stderr, "    -M, --missing CHAR             Output CHAR instead of skipping a missing genotype \"./.\"\n");
+    fprintf(stderr, "    -o, --output FILE              Write output to a file [standard output]\n");
+    fprintf(stderr, "    -p, --prefix STRING            Prefix to add to output sequence names\n");
+    fprintf(stderr, "        --regions-overlap 0|1|2    Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n");
+    fprintf(stderr, "    -s, --samples LIST             Comma-separated list of samples to include, \"-\" to ignore samples and use REF,ALT\n");
+    fprintf(stderr, "    -S, --samples-file FILE        File of samples to include\n");
     fprintf(stderr, "Examples:\n");
     fprintf(stderr, "   # Get the consensus for one region. The fasta header lines are then expected\n");
     fprintf(stderr, "   # in the form \">chr:from-to\".\n");
     fprintf(stderr, "   samtools faidx ref.fa 8:11870-11890 | bcftools consensus in.vcf.gz > out.fa\n");
     fprintf(stderr, "\n");
+    fprintf(stderr, "   # See also http://samtools.github.io/bcftools/howtos/consensus-sequence.html\n");
+    fprintf(stderr, "\n");
     exit(1);
 }
 
@@ -1085,8 +1188,9 @@ int main_consensus(int argc, char *argv[])
 {
     args_t *args = (args_t*) calloc(1,sizeof(args_t));
     args->argc   = argc; args->argv = argv;
+    args->regions_overlap = 1;
 
-    static struct option loptions[] = 
+    static struct option loptions[] =
     {
         {"mark-del",required_argument,NULL,1},
         {"mark-ins",required_argument,NULL,2},
@@ -1095,6 +1199,8 @@ int main_consensus(int argc, char *argv[])
         {"exclude",required_argument,NULL,'e'},
         {"include",required_argument,NULL,'i'},
         {"sample",1,0,'s'},
+        {"samples",1,0,'s'},
+        {"samples-file",1,0,'S'},
         {"iupac-codes",0,0,'I'},
         {"haplotype",1,0,'H'},
         {"output",1,0,'o'},
@@ -1104,32 +1210,40 @@ int main_consensus(int argc, char *argv[])
         {"absent",1,0,'a'},
         {"chain",1,0,'c'},
         {"prefix",required_argument,0,'p'},
+        {"regions-overlap",required_argument,0,5},
         {0,0,0,0}
     };
     int c;
-    while ((c = getopt_long(argc, argv, "h?s:1Ii:e:H:f:o:m:c:M:p:a:",loptions,NULL)) >= 0)
+    while ((c = getopt_long(argc, argv, "h?s:S:1Ii:e:H:f:o:m:c:M:p:a:",loptions,NULL)) >= 0)
     {
-        switch (c) 
+        switch (c)
         {
             case  1 : args->mark_del = optarg[0]; break;
             case  2 :
-                if ( !strcasecmp(optarg,"uc") ) args->mark_ins = 'u';
-                else if ( !strcasecmp(optarg,"lc") ) args->mark_ins = 'l';
+                if ( !strcasecmp(optarg,"uc") ) args->mark_ins = TO_UPPER;
+                else if ( !strcasecmp(optarg,"lc") ) args->mark_ins = TO_LOWER;
+                else if ( !optarg[1] && optarg[0]>32 && optarg[0]<127 ) args->mark_ins = optarg[0];
                 else error("The argument is not recognised: --mark-ins %s\n",optarg);
                 break;
             case  3 :
-                if ( !strcasecmp(optarg,"uc") ) args->mark_snv = 'u';
-                else if ( !strcasecmp(optarg,"lc") ) args->mark_snv = 'l';
+                if ( !strcasecmp(optarg,"uc") ) args->mark_snv = TO_UPPER;
+                else if ( !strcasecmp(optarg,"lc") ) args->mark_snv = TO_LOWER;
+                else if ( !optarg[1] && optarg[0]>32 && optarg[0]<127 ) args->mark_snv = optarg[0];
                 else error("The argument is not recognised: --mark-snv %s\n",optarg);
                 break;
+            case  5 :
+                args->regions_overlap = parse_overlap_option(optarg);
+                if ( args->regions_overlap < 0 ) error("Could not parse: --regions-overlap %s\n",optarg);
+                break;
             case 'p': args->chr_prefix = optarg; break;
             case 's': args->sample = optarg; break;
+            case 'S': args->sample_fname = optarg; break;
             case 'o': args->output_fname = optarg; break;
             case 'I': args->output_iupac = 1; break;
-            case 'e': 
+            case 'e':
                 if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
                 args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
-            case 'i': 
+            case 'i':
                 if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
                 args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
             case 'f': args->ref_fname = optarg; break;
@@ -1139,12 +1253,12 @@ int main_consensus(int argc, char *argv[])
                 args->absent_allele = optarg[0];
                 if ( optarg[1]!=0 ) error("Expected single character with -a, got \"%s\"\n", optarg);
                 break;
-            case 'M': 
-                args->missing_allele = optarg[0]; 
+            case 'M':
+                args->missing_allele = optarg[0];
                 if ( optarg[1]!=0 ) error("Expected single character with -M, got \"%s\"\n", optarg);
                 break;
             case 'c': args->chain_fname = optarg; break;
-            case 'H': 
+            case 'H':
                 if ( !strcasecmp(optarg,"R") ) args->allele |= PICK_REF;
                 else if ( !strcasecmp(optarg,"A") ) args->allele |= PICK_ALT;
                 else if ( !strcasecmp(optarg,"L") ) args->allele |= PICK_LONG|PICK_REF;
@@ -1160,7 +1274,8 @@ int main_consensus(int argc, char *argv[])
                 {
                     char *tmp;
                     args->haplotype = strtol(optarg, &tmp, 10);
-                    if ( tmp==optarg || *tmp ) error("Error: Could not parse --haplotype %s, expected numeric argument\n", optarg);
+                    if ( tmp==optarg || (*tmp && strcasecmp(tmp,"pIu")) ) error("Error: Could not parse \"--haplotype %s\", expected number of number followed with \"pIu\"\n", optarg);
+                    if ( *tmp ) args->allele |= PICK_IUPAC;
                     if ( args->haplotype <=0 ) error("Error: Expected positive integer with --haplotype\n");
                 }
                 break;
