@@ -473,6 +473,10 @@ uint32_t parse_tags(args_t *args, const char *str)
         if ( !strcasecmp(tags[i],"all") )
         {
             flag |= ~(SET_END|SET_TYPE);
+            // include F_MISSING as part of 'all', which requires explicitly
+            // initialising it as a filter expression not just setting a
+            // bitfield flag.
+            flag |= parse_func(args,"F_MISSING=F_MISSING","F_MISSING");
             args->warned = ~(SET_END|SET_TYPE);
             args->unpack |= BCF_UN_FMT;
         }
